from ..graph import DiGraph
import pytest


class ObjTest:
    def __init__(self, name):
        self.name = name


A = ObjTest("a")
B = ObjTest("b")
C = ObjTest("c")
D = ObjTest("d")
E = ObjTest("e")


def test_no_edges():
    """a, b"""
    graph = DiGraph(nodes=[A, B])

    # checking nodes and edges
    assert [nd.name for nd in graph.nodes] == ["a", "b"]
    assert [(edg[0].name, edg[1].name) for edg in graph.edges] == []
    # checking names
    assert set(graph.nodes_names_map.keys()) == {"a", "b"}
    assert graph.edges_names == []


def test_edges_1():
    """a -> b"""
    graph = DiGraph(nodes=[A, B], edges=[(A, B)])

    assert [nd.name for nd in graph.nodes] == ["a", "b"]
    assert [(edg[0].name, edg[1].name) for edg in graph.edges] == [("a", "b")]

    assert set(graph.nodes_names_map.keys()) == {"a", "b"}
    assert graph.edges_names == [("a", "b")]


def test_edges_1a():
    """a -> b (add_nodes and add_edges)"""
    graph = DiGraph()
    graph.add_nodes([A, B])
    graph.add_edges((A, B))

    assert set(graph.nodes_names_map.keys()) == {"a", "b"}
    assert graph.edges_names == [("a", "b")]


def test_edges_2():
    """a -> b"""
    graph = DiGraph(nodes=[B, A], edges=[(A, B)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a"}
    assert graph.edges_names == [("a", "b")]


def test_edges_3():
    """a-> b -> c; a -> c; d"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C), (A, C)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c", "d"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "c")]


def test_edges_ecxeption_1():
    with pytest.raises(Exception) as excinfo:
        graph = DiGraph(nodes=[A, B, A], edges=[(A, B)])
    assert "repeated elements" in str(excinfo.value)


def test_edges_ecxeption_2():
    with pytest.raises(Exception) as excinfo:
        graph = DiGraph(nodes=[A, B], edges=[(A, C)])
    assert "can't be added" in str(excinfo.value)


def test_sort_1():
    """a -> b"""
    graph = DiGraph(nodes=[A, B], edges=[(A, B)])
    assert set(graph.nodes_names_map.keys()) == {"a", "b"}
    assert graph.edges_names == [("a", "b")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "b"]


def test_sort_2():
    """a -> b"""
    graph = DiGraph(nodes=[B, A], edges=[(A, B)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a"}
    assert graph.edges_names == [("a", "b")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "b"]


def test_sort_3():
    """a-> b -> c; a -> c; d"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C), (A, C)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c", "d"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "c")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "d", "b", "c"]


def test_sort_4():
    """a-> b -> c; a -> c; a -> d"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C), (A, C), (A, D)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c", "d"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "c"), ("a", "d")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "b", "d", "c"]


def test_sort_5():
    """a-> b -> c; a -> c; d -> c"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C), (A, C), (D, C)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c", "d"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "c"), ("d", "c")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "d", "b", "c"]


def test_sort_5a():
    """a-> b -> c; a -> c; d -> c (add_nodes/edges)"""
    graph = DiGraph(nodes=[A, C, D], edges=[(A, C), (D, C)])
    graph.add_nodes(B)
    graph.add_edges([(A, B), (B, C)])
    assert set(graph.nodes_names_map.keys()) == {"a", "c", "d", "b"}
    assert graph.edges_names == [("a", "c"), ("d", "c"), ("a", "b"), ("b", "c")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "d", "b", "c"]


def test_sort_5b():
    """a-> b -> c; a -> c; d -> c (add_nodes/edges)"""
    graph = DiGraph(nodes=[A, C, D], edges=[(A, C), (D, C)])
    assert set(graph.nodes_names_map.keys()) == {"a", "c", "d"}
    assert graph.edges_names == [("a", "c"), ("d", "c")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "d", "c"]

    graph.add_nodes(B)
    assert set(graph.nodes_names_map.keys()) == {"a", "c", "d", "b"}
    assert graph.edges_names == [("a", "c"), ("d", "c")]
    assert graph.sorted_nodes_names == ["a", "d", "b", "c"]

    graph.add_edges([(A, B), (B, C)])
    assert set(graph.nodes_names_map.keys()) == {"a", "c", "d", "b"}
    assert graph.edges_names == [("a", "c"), ("d", "c"), ("a", "b"), ("b", "c")]
    assert graph.sorted_nodes_names == ["a", "d", "b", "c"]


def test_sort_6():
    """a -> b -> c -> e; a -> c -> e; a -> b -> d -> e"""
    graph = DiGraph(
        nodes=[D, E, C, B, A], edges=[(A, B), (B, C), (A, C), (B, D), (C, E), (D, E)]
    )
    assert set(graph.nodes_names_map.keys()) == {"d", "e", "c", "b", "a"}
    assert graph.edges_names == [
        ("a", "b"),
        ("b", "c"),
        ("a", "c"),
        ("b", "d"),
        ("c", "e"),
        ("d", "e"),
    ]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "b", "d", "c", "e"]


def test_remove_1():
    """a -> b (removing a node)"""
    graph = DiGraph(nodes=[A, B], edges=[(A, B)])
    assert set(graph.nodes_names_map.keys()) == {"a", "b"}
    assert graph.edges_names == [("a", "b")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "b"]

    # removing a node (e.g. after is sent to run)
    graph.remove_nodes(A)
    assert set(graph.nodes_names_map.keys()) == {"b"}
    assert graph.edges_names == [("a", "b")]
    assert graph.sorted_nodes_names == ["b"]

    # removing all connections (e.g. after the task is done)
    graph.remove_nodes_connections(A)
    assert set(graph.nodes_names_map.keys()) == {"b"}
    assert graph.edges_names == []
    assert graph.sorted_nodes_names == ["b"]


def test_remove_2():
    """a-> b -> c; a -> c; d (removing a node)"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C), (A, C)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c", "d"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "c")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "d", "b", "c"]

    graph.remove_nodes(A)
    assert set(graph.nodes_names_map.keys()) == {"b", "c", "d"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "c")]
    assert graph.sorted_nodes_names == ["d", "b", "c"]

    graph.remove_nodes_connections(A)
    assert set(graph.nodes_names_map.keys()) == {"b", "c", "d"}
    assert graph.edges_names == [("b", "c")]
    assert graph.sorted_nodes_names == ["d", "b", "c"]


def test_remove_3():
    """a-> b -> c; a -> c; d (removing a node)"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C), (A, C)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c", "d"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "c")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "d", "b", "c"]

    graph.remove_nodes(D)
    graph.remove_nodes_connections(D)
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "c")]
    assert graph.sorted_nodes_names == ["a", "b", "c"]


def test_remove_4():
    """ a-> b -> c; a -> d -> e (removing A and later D)"""
    graph = DiGraph(nodes=[B, A, C, D, E], edges=[(A, B), (B, C), (A, D), (D, E)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c", "d", "e"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "d"), ("d", "e")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "b", "d", "c", "e"]

    graph.remove_nodes(A)
    graph.remove_nodes_connections(A)
    assert set(graph.nodes_names_map.keys()) == {"b", "c", "d", "e"}
    assert graph.edges_names == [("b", "c"), ("d", "e")]
    assert graph.sorted_nodes_names == ["b", "d", "c", "e"]

    graph.remove_nodes(D)
    graph.remove_nodes_connections(D)
    assert set(graph.nodes_names_map.keys()) == {"b", "c", "e"}
    assert graph.edges_names == [("b", "c")]
    assert graph.sorted_nodes_names == ["b", "e", "c"]


def test_remove_5():
    """ a-> b -> c; a -> d -> e (removing A, and [B, D] at the same time)"""
    graph = DiGraph(nodes=[B, A, C, D, E], edges=[(A, B), (B, C), (A, D), (D, E)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c", "d", "e"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "d"), ("d", "e")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "b", "d", "c", "e"]

    graph.remove_nodes(A)
    graph.remove_nodes_connections(A)
    assert set(graph.nodes_names_map.keys()) == {"b", "c", "d", "e"}
    assert graph.edges_names == [("b", "c"), ("d", "e")]
    assert graph.sorted_nodes_names == ["b", "d", "c", "e"]

    graph.remove_nodes([B, D])
    graph.remove_nodes_connections([B, D])
    assert set(graph.nodes_names_map.keys()) == {"c", "e"}
    assert graph.edges_names == []
    assert graph.sorted_nodes_names == ["c", "e"]


def test_remove_exception_1():
    """a -> b (removing a node)"""
    graph = DiGraph(nodes=[A, B], edges=[(A, B)])
    assert set(graph.nodes_names_map.keys()) == {"a", "b"}
    assert graph.edges_names == [("a", "b")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "b"]

    with pytest.raises(Exception) as excinfo:
        graph.remove_nodes(B)
    assert "has to wait" in str(excinfo.value)


def test_remove_add_1():
    """a -> b (removing and adding nodes)"""
    graph = DiGraph(nodes=[A, B], edges=[(A, B)])
    assert set(graph.nodes_names_map.keys()) == {"a", "b"}
    assert graph.edges_names == [("a", "b")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "b"]

    # removing a node (e.g. after is sent to run)
    graph.remove_nodes(A)
    graph.remove_nodes_connections(A)
    assert set(graph.nodes_names_map.keys()) == {"b"}
    assert graph.edges_names == []
    assert graph.sorted_nodes_names == ["b"]

    graph.add_nodes(A)
    graph.add_edges((A, B))
    assert set(graph.nodes_names_map.keys()) == {"b", "a"}
    assert graph.edges_names == [("a", "b")]
    assert graph.sorted_nodes_names == ["a", "b"]


def test_remove_add_2():
    """a-> b -> c; a -> c; d (removing and adding nodes)"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C), (A, C)])
    assert set(graph.nodes_names_map.keys()) == {"b", "a", "c", "d"}
    assert graph.edges_names == [("a", "b"), ("b", "c"), ("a", "c")]

    graph.sorting()
    assert graph.sorted_nodes_names == ["a", "d", "b", "c"]

    graph.remove_nodes(A)
    graph.remove_nodes_connections(A)
    assert set(graph.nodes_names_map.keys()) == {"b", "c", "d"}
    assert graph.edges_names == [("b", "c")]
    assert graph.sorted_nodes_names == ["d", "b", "c"]

    graph.add_nodes(A)
    graph.add_edges([(A, B), (A, C)])
    assert set(graph.nodes_names_map.keys()) == {"b", "c", "d", "a"}
    assert graph.edges_names == [("b", "c"), ("a", "b"), ("a", "c")]
    assert graph.sorted_nodes_names == ["d", "a", "b", "c"]


def test_maxpath_1():
    """a -> b"""
    graph = DiGraph(nodes=[B, A], edges=[(A, B)])
    graph.calculate_max_paths()
    assert list(graph.max_paths.keys()) == ["a"]
    assert graph.max_paths["a"] == {"b": 1}


def test_maxpath_2():
    """a -> b -> c"""
    graph = DiGraph(nodes=[B, A, C], edges=[(A, B), (B, C)])
    graph.calculate_max_paths()
    assert list(graph.max_paths.keys()) == ["a"]
    assert graph.max_paths["a"] == {"b": 1, "c": 2}


def test_maxpath_3():
    """a -> b -> c; d"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C)])
    graph.calculate_max_paths()
    assert list(graph.max_paths.keys()) == ["a", "d"]
    assert graph.max_paths["a"] == {"b": 1, "c": 2}
    assert graph.max_paths["d"] == {}


def test_maxpath_4():
    """a-> b -> c; a -> c; d"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C), (A, C)])
    graph.calculate_max_paths()
    assert list(graph.max_paths.keys()) == ["a", "d"]
    assert graph.max_paths["a"] == {"b": 1, "c": 2}
    assert graph.max_paths["d"] == {}


def test_maxpath_5():
    """a-> b -> c; a -> c; d -> b"""
    graph = DiGraph(nodes=[B, A, C, D], edges=[(A, B), (B, C), (A, C), (D, B)])
    graph.calculate_max_paths()
    assert list(graph.max_paths.keys()) == ["a", "d"]
    assert graph.max_paths["a"] == {"b": 1, "c": 2}
    assert graph.max_paths["d"] == {"b": 1, "c": 2}


def test_copy_1():
    """a -> b"""
    graph = DiGraph(nodes=[B, A], edges=[(A, B)])
    graph_copy = graph.copy()

    assert graph.nodes == graph_copy.nodes
    assert id(graph.nodes) != (graph_copy.nodes)
    assert id(graph.nodes[0]) == id(graph_copy.nodes[0])
    assert graph.edges == graph_copy.edges
    assert id(graph.edges) != (graph_copy.edges)



# -*- coding: utf-8 -*-

import os, shutil
import subprocess as sp
import pytest
import attr

from ..task import DockerTask, ShellCommandTask
from ..submitter import Submitter
from ..core import Workflow
from ..specs import ShellOutSpec, SpecInfo, File, DockerSpec

if bool(shutil.which("sbatch")):
    Plugins = ["cf", "slurm"]
else:
    Plugins = ["cf"]

need_docker = pytest.mark.skipif(
    shutil.which("docker") is None or sp.call(["docker", "info"]),
    reason="no docker within the container",
)


@need_docker
def test_docker_1_nosubm():
    """ simple command in a container, a default bindings and working directory is added
        no submitter
    """
    cmd = "whoami"
    docky = DockerTask(name="docky", executable=cmd, image="busybox")
    assert docky.inputs.image == "busybox"
    assert docky.inputs.container == "docker"
    assert (
        docky.cmdline
        == f"docker run --rm -v {docky.output_dir}:/output_pydra:rw -w /output_pydra {docky.inputs.image} {cmd}"
    )

    res = docky()
    assert res.output.stdout == "root\n"
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_1(plugin):
    """ simple command in a container, a default bindings and working directory is added
        using submitter
    """
    cmd = "whoami"
    docky = DockerTask(name="docky", executable=cmd, image="busybox")

    with Submitter(plugin=plugin) as sub:
        docky(submitter=sub)

    res = docky.result()
    assert res.output.stdout == "root\n"
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_1_dockerflag(plugin):
    """ simple command in a container, a default bindings and working directory is added
        using ShellComandTask with container_info=("docker", image)
    """
    cmd = "whoami"
    shocky = ShellCommandTask(
        name="shocky", executable=cmd, container_info=("docker", "busybox")
    )

    with Submitter(plugin=plugin) as sub:
        shocky(submitter=sub)

    res = shocky.result()
    assert res.output.stdout == "root\n"
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_1_dockerflag_exception(plugin):
    """using ShellComandTask with container_info=("docker"), no image provided"""
    cmd = "whoami"
    with pytest.raises(Exception) as excinfo:
        shocky = ShellCommandTask(
            name="shocky", executable=cmd, container_info=("docker")
        )
    assert "container_info has to have 2 or 3 elements" in str(excinfo.value)


@need_docker
def test_docker_2_nosubm():
    """ a command with arguments, cmd and args given as executable
        no submitter
    """
    cmd = ["echo", "hail", "pydra"]
    docky = DockerTask(name="docky", executable=cmd, image="busybox")
    assert (
        docky.cmdline
        == f"docker run --rm -v {docky.output_dir}:/output_pydra:rw -w /output_pydra {docky.inputs.image} {' '.join(cmd)}"
    )

    res = docky()
    assert res.output.stdout.strip() == " ".join(cmd[1:])
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_2(plugin):
    """ a command with arguments, cmd and args given as executable
        using submitter
    """
    cmd = ["echo", "hail", "pydra"]
    docky = DockerTask(name="docky", executable=cmd, image="busybox")
    assert (
        docky.cmdline
        == f"docker run --rm -v {docky.output_dir}:/output_pydra:rw -w /output_pydra {docky.inputs.image} {' '.join(cmd)}"
    )

    with Submitter(plugin=plugin) as sub:
        docky(submitter=sub)
    res = docky.result()
    assert res.output.stdout.strip() == " ".join(cmd[1:])
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_2_dockerflag(plugin):
    """ a command with arguments, cmd and args given as executable
        using ShellComandTask with container_info=("docker", image)
    """
    cmd = ["echo", "hail", "pydra"]
    shocky = ShellCommandTask(
        name="shocky", executable=cmd, container_info=("docker", "busybox")
    )
    assert (
        shocky.cmdline
        == f"docker run --rm -v {shocky.output_dir}:/output_pydra:rw -w /output_pydra {shocky.inputs.image} {' '.join(cmd)}"
    )

    with Submitter(plugin=plugin) as sub:
        shocky(submitter=sub)
    res = shocky.result()
    assert res.output.stdout.strip() == " ".join(cmd[1:])
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
def test_docker_2a_nosubm():
    """ a command with arguments, using executable and args
        no submitter
    """
    cmd_exec = "echo"
    cmd_args = ["hail", "pydra"]
    # separate command into exec + args
    docky = DockerTask(
        name="docky", executable=cmd_exec, args=cmd_args, image="busybox"
    )
    assert docky.inputs.executable == "echo"
    assert (
        docky.cmdline
        == f"docker run --rm -v {docky.output_dir}:/output_pydra:rw -w /output_pydra {docky.inputs.image} {cmd_exec} {' '.join(cmd_args)}"
    )

    res = docky()
    assert res.output.stdout.strip() == " ".join(cmd_args)
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_2a(plugin):
    """ a command with arguments, using executable and args
        using submitter
    """
    cmd_exec = "echo"
    cmd_args = ["hail", "pydra"]
    # separate command into exec + args
    docky = DockerTask(
        name="docky", executable=cmd_exec, args=cmd_args, image="busybox"
    )
    assert docky.inputs.executable == "echo"
    assert (
        docky.cmdline
        == f"docker run --rm -v {docky.output_dir}:/output_pydra:rw -w /output_pydra {docky.inputs.image} {cmd_exec} {' '.join(cmd_args)}"
    )

    with Submitter(plugin=plugin) as sub:
        docky(submitter=sub)
    res = docky.result()
    assert res.output.stdout.strip() == " ".join(cmd_args)
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_3(plugin, tmpdir):
    """ a simple command in container with bindings,
        creating directory in tmp dir and checking if it is in the container
    """
    # creating a new directory
    tmpdir.mkdir("new_dir")
    cmd = ["ls", "/tmp_dir"]
    docky = DockerTask(name="docky", executable=cmd, image="busybox")
    # binding tmp directory to the container
    docky.inputs.bindings = [(str(tmpdir), "/tmp_dir", "ro")]

    with Submitter(plugin=plugin) as sub:
        docky(submitter=sub)

    res = docky.result()
    assert res.output.stdout == "new_dir\n"
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_3_dockerflag(plugin, tmpdir):
    """ a simple command in container with bindings,
        creating directory in tmp dir and checking if it is in the container
        using ShellComandTask with container_info=("docker", image)
    """
    # creating a new directory
    tmpdir.mkdir("new_dir")
    cmd = ["ls", "/tmp_dir"]
    shocky = ShellCommandTask(
        name="shocky", container_info=("docker", "busybox"), executable=cmd
    )
    # binding tmp directory to the container
    shocky.inputs.bindings = [(str(tmpdir), "/tmp_dir", "ro")]

    with Submitter(plugin=plugin) as sub:
        shocky(submitter=sub)

    res = shocky.result()
    assert res.output.stdout == "new_dir\n"
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_3_dockerflagbind(plugin, tmpdir):
    """ a simple command in container with bindings,
        creating directory in tmp dir and checking if it is in the container
        using ShellComandTask with container_info=("docker", image)
    """
    # creating a new directory
    tmpdir.mkdir("new_dir")
    cmd = ["ls", "/tmp_dir"]
    shocky = ShellCommandTask(
        name="shocky",
        container_info=("docker", "busybox", [(str(tmpdir), "/tmp_dir", "ro")]),
        executable=cmd,
    )

    with Submitter(plugin=plugin) as sub:
        shocky(submitter=sub)

    res = shocky.result()
    assert res.output.stdout == "new_dir\n"
    assert res.output.return_code == 0
    if res.output.stderr:
        assert "Unable to find image" in res.output.stderr


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_4(plugin, tmpdir):
    """ task reads the file that is bounded to the container
        specifying bindings,
    """
    with open(tmpdir.join("file_pydra.txt"), "w") as f:
        f.write("hello from pydra")

    cmd = ["cat", "/tmp_dir/file_pydra.txt"]
    docky = DockerTask(
        name="docky_cat",
        image="busybox",
        executable=cmd,
        bindings=[(str(tmpdir), "/tmp_dir", "ro")],
        strip=True,
    )

    with Submitter(plugin=plugin) as sub:
        docky(submitter=sub)

    res = docky.result()
    assert res.output.stdout == "hello from pydra"
    assert res.output.return_code == 0


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_4_dockerflag(plugin, tmpdir):
    """ task reads the file that is bounded to the container
        specifying bindings,
        using ShellComandTask with container_info=("docker", image, bindings)
    """
    with open(tmpdir.join("file_pydra.txt"), "w") as f:
        f.write("hello from pydra")

    cmd = ["cat", "/tmp_dir/file_pydra.txt"]
    shocky = ShellCommandTask(
        name="shocky",
        container_info=("docker", "busybox", [(str(tmpdir), "/tmp_dir", "ro")]),
        executable=cmd,
        strip=True,
    )

    with Submitter(plugin=plugin) as sub:
        shocky(submitter=sub)

    res = shocky.result()
    assert res.output.stdout == "hello from pydra"
    assert res.output.return_code == 0


# tests with State


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_st_1(plugin):
    """ commands without arguments in container
        splitter = executable
    """
    cmd = ["pwd", "whoami"]
    docky = DockerTask(name="docky", executable=cmd, image="busybox").split(
        "executable"
    )
    assert docky.state.splitter == "docky.executable"

    for ii, el in enumerate(docky.cmdline):
        assert (
            el
            == f"docker run --rm -v {docky.output_dir[ii]}:/output_pydra:rw -w /output_pydra {docky.inputs.image} {cmd[ii]}"
        )

    res = docky(plugin=plugin)
    assert res[0].output.stdout == "/output_pydra\n"
    assert res[1].output.stdout == "root\n"
    assert res[0].output.return_code == res[1].output.return_code == 0


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_st_2(plugin):
    """ command with arguments in docker, checking the distribution
        splitter = image
    """
    cmd = ["cat", "/etc/issue"]
    docky = DockerTask(name="docky", executable=cmd, image=["debian", "ubuntu"]).split(
        "image"
    )
    assert docky.state.splitter == "docky.image"

    for ii, el in enumerate(docky.cmdline):
        assert (
            el
            == f"docker run --rm -v {docky.output_dir[ii]}:/output_pydra:rw -w /output_pydra {docky.inputs.image[ii]} {' '.join(cmd)}"
        )

    res = docky(plugin=plugin)
    assert "Debian" in res[0].output.stdout
    assert "Ubuntu" in res[1].output.stdout
    assert res[0].output.return_code == res[1].output.return_code == 0


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_st_3(plugin):
    """ outer splitter image and executable
    """
    cmd = ["whoami", ["cat", "/etc/issue"]]
    docky = DockerTask(name="docky", executable=cmd, image=["debian", "ubuntu"]).split(
        ["image", "executable"]
    )
    assert docky.state.splitter == ["docky.image", "docky.executable"]
    res = docky(plugin=plugin)

    assert res[0].output.stdout == "root\n"
    assert "Debian" in res[1].output.stdout
    assert res[2].output.stdout == "root\n"
    assert "Ubuntu" in res[3].output.stdout


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_st_4(plugin):
    """ outer splitter image and executable, combining with images
    """
    cmd = ["whoami", ["cat", "/etc/issue"]]
    docky = (
        DockerTask(name="docky", executable=cmd, image=["debian", "ubuntu"])
        .split(["image", "executable"])
        .combine("image")
    )
    assert docky.state.splitter == ["docky.image", "docky.executable"]
    assert docky.state.combiner == ["docky.image"]
    assert docky.state.splitter_final == "docky.executable"

    for ii, el in enumerate(docky.cmdline):
        i, j = ii // 2, ii % 2
        if j == 0:
            cmd_str = "whoami"
        else:
            cmd_str = " ".join(["cat", "/etc/issue"])
        assert (
            el
            == f"docker run --rm -v {docky.output_dir[ii]}:/output_pydra:rw -w /output_pydra {docky.inputs.image[i]} {cmd_str}"
        )

    res = docky(plugin=plugin)

    # checking the first command
    res_cmd1 = res[0]
    assert res_cmd1[0].output.stdout == "root\n"
    assert res_cmd1[1].output.stdout == "root\n"

    # checking the second command
    res_cmd2 = res[1]
    assert "Debian" in res_cmd2[0].output.stdout
    assert "Ubuntu" in res_cmd2[1].output.stdout


# tests with workflows


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_wf_docker_1(plugin, tmpdir):
    """ a workflow with two connected task
        the first one read the file that is bounded to the container,
        the second uses echo
    """
    with open(tmpdir.join("file_pydra.txt"), "w") as f:
        f.write("hello from pydra")

    wf = Workflow(name="wf", input_spec=["cmd1", "cmd2"])
    wf.inputs.cmd1 = ["cat", "/tmp_dir/file_pydra.txt"]
    wf.inputs.cmd2 = ["echo", "message from the previous task:"]
    wf.add(
        DockerTask(
            name="docky_cat",
            image="busybox",
            executable=wf.lzin.cmd1,
            bindings=[(str(tmpdir), "/tmp_dir", "ro")],
            strip=True,
        )
    )
    wf.add(
        DockerTask(
            name="docky_echo",
            image="ubuntu",
            executable=wf.lzin.cmd2,
            args=wf.docky_cat.lzout.stdout,
            strip=True,
        )
    )
    wf.set_output([("out", wf.docky_echo.lzout.stdout)])

    with pytest.raises(Exception) as excinfo:
        wf.docky_echo.cmdline
    assert "can't return cmdline" in str(excinfo.value)

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == "message from the previous task: hello from pydra"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_wf_docker_1_dockerflag(plugin, tmpdir):
    """ a workflow with two connected task
        the first one read the file that is bounded to the container,
        the second uses echo
        using ShellComandTask with container_info
    """
    with open(tmpdir.join("file_pydra.txt"), "w") as f:
        f.write("hello from pydra")

    wf = Workflow(name="wf", input_spec=["cmd1", "cmd2"])
    wf.inputs.cmd1 = ["cat", "/tmp_dir/file_pydra.txt"]
    wf.inputs.cmd2 = ["echo", "message from the previous task:"]
    wf.add(
        ShellCommandTask(
            name="shocky_cat",
            container_info=("docker", "busybox", [(str(tmpdir), "/tmp_dir", "ro")]),
            executable=wf.lzin.cmd1,
            strip=True,
        )
    )
    wf.add(
        ShellCommandTask(
            name="shocky_echo",
            executable=wf.lzin.cmd2,
            args=wf.shocky_cat.lzout.stdout,
            strip=True,
            container_info=("docker", "ubuntu"),
        )
    )
    wf.set_output([("out", wf.shocky_echo.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == "message from the previous task: hello from pydra"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_wf_docker_2pre(plugin, tmpdir):
    """ a workflow with two connected task that run python scripts
        the first one creates a text file and the second one reads the file
    """

    scripts_dir = os.path.join(os.path.dirname(__file__), "data_tests")

    cmd1 = ["python", "/scripts/saving.py", "-f", "/outputs/tmp.txt"]
    dt = DockerTask(
        name="save",
        image="python:3.7-alpine",
        executable=cmd1,
        bindings=[(str(tmpdir), "/outputs"), (scripts_dir, "/scripts", "ro")],
        strip=True,
    )
    res = dt(plugin=plugin)
    assert res.output.stdout == "/outputs/tmp.txt"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_wf_docker_2(plugin, tmpdir):
    """ a workflow with two connected task that run python scripts
        the first one creates a text file and the second one reads the file
    """

    scripts_dir = os.path.join(os.path.dirname(__file__), "data_tests")

    wf = Workflow(name="wf", input_spec=["cmd1", "cmd2"])
    wf.inputs.cmd1 = ["python", "/scripts/saving.py", "-f", "/outputs/tmp.txt"]
    wf.inputs.cmd2 = ["python", "/scripts/loading.py", "-f"]
    wf.add(
        DockerTask(
            name="save",
            image="python:3.7-alpine",
            executable=wf.lzin.cmd1,
            bindings=[(str(tmpdir), "/outputs"), (scripts_dir, "/scripts", "ro")],
            strip=True,
        )
    )
    wf.add(
        DockerTask(
            name="load",
            image="python:3.7-alpine",
            executable=wf.lzin.cmd2,
            args=wf.save.lzout.stdout,
            bindings=[(str(tmpdir), "/outputs"), (scripts_dir, "/scripts", "ro")],
            strip=True,
        )
    )
    wf.set_output([("out", wf.load.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == "Hello!"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_wf_docker_3(plugin, tmpdir):
    """ a workflow with two connected task
        the first one read the file that contains the name of the image,
        the output is passed to the second task as the image used to run the task
    """
    with open(tmpdir.join("image.txt"), "w") as f:
        f.write("ubuntu")

    wf = Workflow(name="wf", input_spec=["cmd1", "cmd2"])
    wf.inputs.cmd1 = ["cat", "/tmp_dir/image.txt"]
    wf.inputs.cmd2 = ["echo", "image passed to the second task:"]
    wf.add(
        DockerTask(
            name="docky_cat",
            image="busybox",
            executable=wf.lzin.cmd1,
            bindings=[(str(tmpdir), "/tmp_dir", "ro")],
            strip=True,
        )
    )
    wf.add(
        DockerTask(
            name="docky_echo",
            image=wf.docky_cat.lzout.stdout,
            executable=wf.lzin.cmd2,
            args=wf.docky_cat.lzout.stdout,
            strip=True,
        )
    )
    wf.set_output([("out", wf.docky_echo.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == "image passed to the second task: ubuntu"


# tests with customized output_spec


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_outputspec_1(plugin, tmpdir):
    """
        customised output_spec, adding files to the output, providing specific pathname
        output_path is automatically added to the bindings
    """
    cmd = ["touch", "newfile_tmp.txt"]
    my_output_spec = SpecInfo(
        name="Output",
        fields=[("newfile", File, "newfile_tmp.txt")],
        bases=(ShellOutSpec,),
    )
    docky = DockerTask(
        name="docky", image="ubuntu", executable=cmd, output_spec=my_output_spec
    )

    with Submitter(plugin=plugin) as sub:
        docky(submitter=sub)

    res = docky.result()
    assert res.output.stdout == ""
    assert res.output.newfile.exists()


# tests with customised input_spec


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_inputspec_1(plugin, tmpdir):
    """ a simple customized input spec for docker task """
    filename = str(tmpdir.join("file_pydra.txt"))
    with open(filename, "w") as f:
        f.write("hello from pydra")

    cmd = "cat"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        file=filename,
        input_spec=my_input_spec,
        strip=True,
    )

    res = docky()
    assert res.output.stdout == "hello from pydra"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_inputspec_1a(plugin, tmpdir):
    """ a simple customized input spec for docker task
        a default value is used
    """
    filename = str(tmpdir.join("file_pydra.txt"))
    with open(filename, "w") as f:
        f.write("hello from pydra")

    cmd = "cat"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    default=filename,
                    metadata={"position": 1, "help_string": "input file"},
                ),
            )
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        input_spec=my_input_spec,
        strip=True,
    )

    res = docky()
    assert res.output.stdout == "hello from pydra"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_inputspec_2(plugin, tmpdir):
    """ a customized input spec with two fields for docker task """
    filename_1 = tmpdir.join("file_pydra.txt")
    with open(filename_1, "w") as f:
        f.write("hello from pydra\n")

    filename_2 = tmpdir.join("file_nice.txt")
    with open(filename_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file1",
                attr.ib(
                    type=File, metadata={"position": 1, "help_string": "input file 1"}
                ),
            ),
            (
                "file2",
                attr.ib(
                    type=File,
                    default=filename_2,
                    metadata={"position": 2, "help_string": "input file 2"},
                ),
            ),
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        file1=filename_1,
        input_spec=my_input_spec,
        strip=True,
    )

    res = docky()
    assert res.output.stdout == "hello from pydra\nhave a nice one"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_inputspec_2a_except(plugin, tmpdir):
    """ a customized input spec with two fields
        first one uses a default, and second doesn't - raises a dataclass exception
    """
    filename_1 = tmpdir.join("file_pydra.txt")
    with open(filename_1, "w") as f:
        f.write("hello from pydra\n")
    filename_2 = tmpdir.join("file_nice.txt")
    with open(filename_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"

    # the field with default value can't be before value without default
    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file1",
                attr.ib(
                    type=File,
                    default=filename_1,
                    metadata={"position": 1, "help_string": "input file 1"},
                ),
            ),
            (
                "file2",
                attr.ib(
                    type=File, metadata={"position": 2, "help_string": "input file 2"}
                ),
            ),
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        file2=filename_2,
        input_spec=my_input_spec,
        strip=True,
    )
    assert docky.inputs.file2 == filename_2

    res = docky()
    assert res.output.stdout == "hello from pydra\nhave a nice one"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_inputspec_2a(plugin, tmpdir):
    """ a customized input spec with two fields
        first one uses a default value
        this is fine even if the second field is not using any defaults
    """
    filename_1 = tmpdir.join("file_pydra.txt")
    with open(filename_1, "w") as f:
        f.write("hello from pydra\n")
    filename_2 = tmpdir.join("file_nice.txt")
    with open(filename_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"

    # if you want set default in the first field you can use default value
    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file1",
                attr.ib(
                    type=File,
                    default=filename_1,
                    metadata={"position": 1, "help_string": "input file 1"},
                ),
            ),
            (
                "file2",
                attr.ib(
                    type=File, metadata={"position": 2, "help_string": "input file 2"}
                ),
            ),
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        file2=filename_2,
        input_spec=my_input_spec,
        strip=True,
    )

    res = docky()
    assert res.output.stdout == "hello from pydra\nhave a nice one"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_inputspec_3(plugin, tmpdir):
    """ input file is in the container, so metadata["container_path"]: True,
        the input will be treated as a str """
    filename = "/proc/1/cgroup"

    cmd = "cat"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                        "container_path": True,
                    },
                ),
            )
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        file=filename,
        input_spec=my_input_spec,
        strip=True,
    )

    cmdline = docky.cmdline
    res = docky()
    assert "docker" in res.output.stdout
    assert cmdline == docky.cmdline


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_inputspec_3a(plugin, tmpdir):
    """ input file does not exist in the local file system,
        but metadata["container_path"] is not used,
        so exception is raised
    """
    filename = "/_proc/1/cgroup"

    cmd = "cat"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        file=filename,
        input_spec=my_input_spec,
        strip=True,
    )

    with pytest.raises(Exception) as excinfo:
        res = docky()
    assert "use field.metadata['container_path']=True" in str(excinfo.value)


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_cmd_inputspec_copyfile_1(plugin, tmpdir):
    """ shelltask changes a file in place,
        adding copyfile=True to the file-input from input_spec
        hardlink or copy in the output_dir should be created
    """
    file = tmpdir.join("file_pydra.txt")
    with open(file, "w") as f:
        f.write("hello from pydra\n")

    cmd = ["sed", "-is", "s/hello/hi/"]

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "orig_file",
                attr.ib(
                    type=File,
                    metadata={
                        "position": 1,
                        "help_string": "orig file",
                        "mandatory": True,
                        "copyfile": True,
                    },
                ),
            ),
            (
                "out_file",
                attr.ib(
                    type=str,
                    metadata={
                        "output_file_template": "{orig_file}",
                        "help_string": "output file",
                    },
                ),
            ),
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        input_spec=my_input_spec,
        orig_file=str(file),
    )

    res = docky()
    assert res.output.stdout == ""
    assert res.output.out_file.exists()
    # the file is  copied, and than it is changed in place
    assert res.output.out_file.parent == docky.output_dir
    with open(res.output.out_file, "r") as f:
        assert "hi from pydra\n" == f.read()
    # the original file is unchanged
    with open(file, "r") as f:
        assert "hello from pydra\n" == f.read()


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_inputspec_state_1(plugin, tmpdir):
    """ a customised input spec for a docker file with a splitter,
        splitter is on files
    """
    filename_1 = tmpdir.join("file_pydra.txt")
    with open(filename_1, "w") as f:
        f.write("hello from pydra\n")
    filename_2 = tmpdir.join("file_nice.txt")
    with open(filename_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    filename = [str(filename_1), str(filename_2)]

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        file=filename,
        input_spec=my_input_spec,
        strip=True,
    ).split("file")

    res = docky()
    assert res[0].output.stdout == "hello from pydra"
    assert res[1].output.stdout == "have a nice one"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_inputspec_state_1b(plugin, tmpdir):
    """ a customised input spec for a docker file with a splitter,
        files from the input spec have the same path in the local os and the container,
        so hash is calculated and the test works fine
    """
    file_1 = tmpdir.join("file_pydra.txt")
    file_2 = tmpdir.join("file_nice.txt")
    with open(file_1, "w") as f:
        f.write("hello from pydra")
    with open(file_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    filename = [str(file_1), str(file_2)]

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(DockerSpec,),
    )

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=cmd,
        file=filename,
        input_spec=my_input_spec,
        strip=True,
    ).split("file")

    res = docky()
    assert res[0].output.stdout == "hello from pydra"
    assert res[1].output.stdout == "have a nice one"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_wf_inputspec_1(plugin, tmpdir):
    """ a customized input spec for workflow with docker tasks """
    filename = tmpdir.join("file_pydra.txt")
    with open(filename, "w") as f:
        f.write("hello from pydra")

    cmd = "cat"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(DockerSpec,),
    )

    wf = Workflow(name="wf", input_spec=["cmd", "file"])
    wf.inputs.cmd = cmd
    wf.inputs.file = filename

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=wf.lzin.cmd,
        file=wf.lzin.file,
        input_spec=my_input_spec,
        strip=True,
    )
    wf.add(docky)

    wf.set_output([("out", wf.docky.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == "hello from pydra"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_wf_state_inputspec_1(plugin, tmpdir):
    """ a customized input spec for workflow with docker tasks that has a state"""
    file_1 = tmpdir.join("file_pydra.txt")
    file_2 = tmpdir.join("file_nice.txt")
    with open(file_1, "w") as f:
        f.write("hello from pydra")
    with open(file_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    filename = [str(file_1), str(file_2)]

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(DockerSpec,),
    )

    wf = Workflow(name="wf", input_spec=["cmd", "file"])
    wf.inputs.cmd = cmd
    wf.inputs.file = filename

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=wf.lzin.cmd,
        file=wf.lzin.file,
        input_spec=my_input_spec,
        strip=True,
    )
    wf.add(docky)
    wf.split("file")

    wf.set_output([("out", wf.docky.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res[0].output.out == "hello from pydra"
    assert res[1].output.out == "have a nice one"


@need_docker
@pytest.mark.parametrize("plugin", Plugins)
def test_docker_wf_ndst_inputspec_1(plugin, tmpdir):
    """ a customized input spec for workflow with docker tasks with states"""
    file_1 = tmpdir.join("file_pydra.txt")
    file_2 = tmpdir.join("file_nice.txt")
    with open(file_1, "w") as f:
        f.write("hello from pydra")
    with open(file_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    filename = [str(file_1), str(file_2)]

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(DockerSpec,),
    )

    wf = Workflow(name="wf", input_spec=["cmd", "file"])
    wf.inputs.cmd = cmd
    wf.inputs.file = filename

    docky = DockerTask(
        name="docky",
        image="busybox",
        executable=wf.lzin.cmd,
        file=wf.lzin.file,
        input_spec=my_input_spec,
        strip=True,
    ).split("file")
    wf.add(docky)

    wf.set_output([("out", wf.docky.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == ["hello from pydra", "have a nice one"]



""" Decorators to apply to functions used in Pydra workflows """


def annotate(annotation):
    """
    Update the annotation of a function.

    Example
    -------
    >>> import pydra
    >>> @pydra.mark.annotate({'a': int, 'return': float})
    ... def square(a):
    ...     return a ** 2.0

    """
    import inspect

    def decorate(func):
        sig = inspect.signature(func)
        unknown = set(annotation) - set(sig.parameters) - {"return"}
        if unknown:
            raise TypeError(f"Cannot annotate unknown parameters: {tuple(unknown)}")
        func.__annotations__.update(annotation)
        return func

    return decorate


def task(func):
    """
    Promote a function to a :class:`~pydra.engine.task.FunctionTask`.

    Example
    -------
    >>> import pydra
    >>> @pydra.mark.task
    ... def square(a: int) -> float:
    ...     return a ** 2.0

    """
    from ..engine.task import FunctionTask

    def decorate(**kwargs):
        return FunctionTask(func=func, **kwargs)

    return decorate



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Pydra: Dataflow Engine

"""
import sys
from setuptools import setup
import versioneer

# Give setuptools a hint to complain if it's too old a version
# 30.3.0 allows us to put most metadata in setup.cfg
# Should match pyproject.toml
SETUP_REQUIRES = ["setuptools >= 30.3.0"]
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ["wheel"] if "bdist_wheel" in sys.argv else []

if __name__ == "__main__":
    setup(
        name="pydra",
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        setup_requires=SETUP_REQUIRES,
    )



# -*- coding: utf-8 -*-

import os, shutil
import subprocess as sp
import pytest
import attr

from ..task import SingularityTask, DockerTask, ShellCommandTask
from ..submitter import Submitter
from ..core import Workflow
from ..specs import ShellOutSpec, SpecInfo, File, SingularitySpec

if bool(shutil.which("sbatch")):
    Plugins = ["cf", "slurm"]
else:
    Plugins = ["cf"]


need_docker = pytest.mark.skipif(
    shutil.which("docker") is None or sp.call(["docker", "info"]),
    reason="no docker available",
)
need_singularity = pytest.mark.skipif(
    shutil.which("singularity") is None, reason="no singularity available"
)


@need_singularity
def test_singularity_1_nosubm(tmpdir):
    """ simple command in a container, a default bindings and working directory is added
        no submitter
    """
    cmd = "pwd"
    image = "library://sylabsed/linux/alpine"
    singu = SingularityTask(name="singu", executable=cmd, image=image, cache_dir=tmpdir)
    assert singu.inputs.image == "library://sylabsed/linux/alpine"
    assert singu.inputs.container == "singularity"
    assert (
        singu.cmdline
        == f"singularity exec -B {singu.output_dir}:/output_pydra:rw {image} {cmd}"
    )

    res = singu()
    assert "SingularityTask" in res.output.stdout
    assert res.output.return_code == 0


@need_singularity
def test_singularity_2_nosubm(tmpdir):
    """ a command with arguments, cmd and args given as executable
        no submitter
    """
    cmd = ["echo", "hail", "pydra"]
    image = "library://sylabsed/linux/alpine"
    singu = SingularityTask(name="singu", executable=cmd, image=image, cache_dir=tmpdir)
    assert (
        singu.cmdline
        == f"singularity exec -B {singu.output_dir}:/output_pydra:rw {image} {' '.join(cmd)}"
    )

    res = singu()
    assert res.output.stdout.strip() == " ".join(cmd[1:])
    assert res.output.return_code == 0


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_2(plugin, tmpdir):
    """ a command with arguments, cmd and args given as executable
        using submitter
    """
    cmd = ["echo", "hail", "pydra"]
    image = "library://sylabsed/linux/alpine"
    singu = SingularityTask(name="singu", executable=cmd, image=image, cache_dir=tmpdir)
    assert (
        singu.cmdline
        == f"singularity exec -B {singu.output_dir}:/output_pydra:rw {image} {' '.join(cmd)}"
    )

    with Submitter(plugin=plugin) as sub:
        singu(submitter=sub)
    res = singu.result()
    assert res.output.stdout.strip() == " ".join(cmd[1:])
    assert res.output.return_code == 0


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_2_singuflag(plugin, tmpdir):
    """ a command with arguments, cmd and args given as executable
        using ShellComandTask with container_info=("singularity", image)
    """
    cmd = ["echo", "hail", "pydra"]
    image = "library://sylabsed/linux/alpine"
    shingu = ShellCommandTask(
        name="shingu",
        executable=cmd,
        container_info=("singularity", image),
        cache_dir=tmpdir,
    )
    assert (
        shingu.cmdline
        == f"singularity exec -B {shingu.output_dir}:/output_pydra:rw {image} {' '.join(cmd)}"
    )

    with Submitter(plugin=plugin) as sub:
        shingu(submitter=sub)
    res = shingu.result()
    assert res.output.stdout.strip() == " ".join(cmd[1:])
    assert res.output.return_code == 0


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_2a(plugin, tmpdir):
    """ a command with arguments, using executable and args
        using submitter
    """
    cmd_exec = "echo"
    cmd_args = ["hail", "pydra"]
    # separate command into exec + args
    image = "library://sylabsed/linux/alpine"
    singu = SingularityTask(
        name="singu", executable=cmd_exec, args=cmd_args, image=image, cache_dir=tmpdir
    )
    assert (
        singu.cmdline
        == f"singularity exec -B {singu.output_dir}:/output_pydra:rw {image} {cmd_exec} {' '.join(cmd_args)}"
    )

    with Submitter(plugin=plugin) as sub:
        singu(submitter=sub)
    res = singu.result()
    assert res.output.stdout.strip() == " ".join(cmd_args)
    assert res.output.return_code == 0


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_3(plugin, tmpdir):
    """ a simple command in container with bindings,
        creating directory in tmp dir and checking if it is in the container
    """
    # creating a new directory
    tmpdir.mkdir("new_dir")
    cmd = ["ls", "/tmp_dir"]
    image = "library://sylabsed/linux/alpine"
    singu = SingularityTask(name="singu", executable=cmd, image=image, cache_dir=tmpdir)
    # binding tmp directory to the container
    singu.inputs.bindings = [(str(tmpdir), "/tmp_dir", "ro")]

    with Submitter(plugin=plugin) as sub:
        singu(submitter=sub)

    res = singu.result()
    assert "new_dir\n" in res.output.stdout
    assert res.output.return_code == 0


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_3_singuflag(plugin, tmpdir):
    """ a simple command in container with bindings,
        creating directory in tmp dir and checking if it is in the container
        using ShellComandTask with container_info=("singularity", image)
    """
    # creating a new directory
    tmpdir.mkdir("new_dir")
    cmd = ["ls", "/tmp_dir"]
    image = "library://sylabsed/linux/alpine"
    shingu = SingularityTask(
        name="singu",
        executable=cmd,
        container_info=("singularity", image),
        cache_dir=tmpdir,
    )
    # binding tmp directory to the container
    shingu.inputs.bindings = [(str(tmpdir), "/tmp_dir", "ro")]

    with Submitter(plugin=plugin) as sub:
        shingu(submitter=sub)

    res = shingu.result()
    assert "new_dir\n" in res.output.stdout
    assert res.output.return_code == 0


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_3_singuflagbind(plugin, tmpdir):
    """ a simple command in container with bindings,
        creating directory in tmp dir and checking if it is in the container
        using ShellComandTask with container_info=("singularity", image, bindings)
    """
    # creating a new directory
    tmpdir.mkdir("new_dir")
    cmd = ["ls", "/tmp_dir"]
    image = "library://sylabsed/linux/alpine"
    shingu = SingularityTask(
        name="singu",
        executable=cmd,
        container_info=("singularity", image, [(str(tmpdir), "/tmp_dir", "ro")]),
        cache_dir=tmpdir,
    )

    with Submitter(plugin=plugin) as sub:
        shingu(submitter=sub)

    res = shingu.result()
    assert "new_dir\n" in res.output.stdout
    assert res.output.return_code == 0


# tests with State


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_st_1(plugin, tmpdir):
    """ commands without arguments in container
        splitter = executable
    """
    cmd = ["pwd", "ls"]
    image = "library://sylabsed/linux/alpine"
    singu = SingularityTask(
        name="singu", executable=cmd, image=image, cache_dir=tmpdir
    ).split("executable")
    assert singu.state.splitter == "singu.executable"

    res = singu(plugin=plugin)
    assert "SingularityTask" in res[0].output.stdout
    assert res[1].output.stdout == ""
    assert res[0].output.return_code == res[1].output.return_code == 0


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_st_2(plugin, tmpdir):
    """ command with arguments in docker, checking the distribution
        splitter = image
    """
    cmd = ["cat", "/etc/issue"]
    image = ["library://sylabsed/linux/alpine", "library://sylabsed/examples/lolcow"]
    singu = SingularityTask(
        name="singu", executable=cmd, image=image, cache_dir=tmpdir
    ).split("image")
    assert singu.state.splitter == "singu.image"

    res = singu(plugin=plugin)
    assert "Alpine" in res[0].output.stdout
    assert "Ubuntu" in res[1].output.stdout
    assert res[0].output.return_code == res[1].output.return_code == 0


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_st_3(plugin, tmpdir):
    """ outer splitter image and executable
    """
    cmd = ["pwd", ["cat", "/etc/issue"]]
    image = ["library://sylabsed/linux/alpine", "library://sylabsed/examples/lolcow"]
    singu = SingularityTask(
        name="singu", executable=cmd, image=image, cache_dir=tmpdir
    ).split(["image", "executable"])
    assert singu.state.splitter == ["singu.image", "singu.executable"]
    res = singu(plugin=plugin)

    assert "SingularityTask" in res[0].output.stdout
    assert "Alpine" in res[1].output.stdout
    assert "SingularityTask" in res[2].output.stdout
    assert "Ubuntu" in res[3].output.stdout


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_wf_singularity_1(plugin, tmpdir):
    """ a workflow with two connected task
        the first one read the file that is bounded to the container,
        the second uses echo
    """
    with open(tmpdir.join("file_pydra.txt"), "w") as f:
        f.write("hello from pydra")

    image = "library://sylabsed/linux/alpine"
    wf = Workflow(name="wf", input_spec=["cmd1", "cmd2"], cache_dir=tmpdir)
    wf.inputs.cmd1 = ["cat", "/tmp_dir/file_pydra.txt"]
    wf.inputs.cmd2 = ["echo", "message from the previous task:"]
    wf.add(
        SingularityTask(
            name="singu_cat",
            image=image,
            executable=wf.lzin.cmd1,
            bindings=[(str(tmpdir), "/tmp_dir", "ro")],
            strip=True,
        )
    )
    wf.add(
        SingularityTask(
            name="singu_echo",
            image=image,
            executable=wf.lzin.cmd2,
            args=wf.singu_cat.lzout.stdout,
            strip=True,
        )
    )
    wf.set_output([("out", wf.singu_echo.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == "message from the previous task: hello from pydra"


@need_docker
@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_wf_singularity_1a(plugin, tmpdir):
    """ a workflow with two connected task - using both containers: Docker and Singul.
        the first one read the file that is bounded to the container,
        the second uses echo
    """
    with open(tmpdir.join("file_pydra.txt"), "w") as f:
        f.write("hello from pydra")

    image_sing = "library://sylabsed/linux/alpine"
    image_doc = "ubuntu"
    wf = Workflow(name="wf", input_spec=["cmd1", "cmd2"], cache_dir=tmpdir)
    wf.inputs.cmd1 = ["cat", "/tmp_dir/file_pydra.txt"]
    wf.inputs.cmd2 = ["echo", "message from the previous task:"]
    wf.add(
        SingularityTask(
            name="singu_cat",
            image=image_sing,
            executable=wf.lzin.cmd1,
            bindings=[(str(tmpdir), "/tmp_dir", "ro")],
            strip=True,
        )
    )
    wf.add(
        DockerTask(
            name="singu_echo",
            image=image_doc,
            executable=wf.lzin.cmd2,
            args=wf.singu_cat.lzout.stdout,
            strip=True,
        )
    )
    wf.set_output([("out", wf.singu_echo.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == "message from the previous task: hello from pydra"


# tests with customized output_spec


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_outputspec_1(plugin, tmpdir):
    """
        customised output_spec, adding files to the output, providing specific pathname
        output_path is automatically added to the bindings
    """
    cmd = ["touch", "newfile_tmp.txt"]
    image = "library://sylabsed/linux/alpine"

    my_output_spec = SpecInfo(
        name="Output",
        fields=[("newfile", File, "newfile_tmp.txt")],
        bases=(ShellOutSpec,),
    )
    singu = SingularityTask(
        name="singu",
        image=image,
        executable=cmd,
        output_spec=my_output_spec,
        cache_dir=tmpdir,
    )

    with Submitter(plugin=plugin) as sub:
        singu(submitter=sub)

    res = singu.result()
    assert res.output.stdout == ""
    assert res.output.newfile.exists()


# tests with customised input_spec


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_inputspec_1(plugin, tmpdir):
    """ a simple customized input spec for singularity task """
    filename = str(tmpdir.join("file_pydra.txt"))
    with open(filename, "w") as f:
        f.write("hello from pydra")

    cmd = "cat"
    image = "library://sylabsed/linux/alpine"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(SingularitySpec,),
    )

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=cmd,
        file=filename,
        input_spec=my_input_spec,
        strip=True,
        cache_dir=tmpdir,
    )

    res = singu()
    assert res.output.stdout == "hello from pydra"


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_inputspec_1a(plugin, tmpdir):
    """ a simple customized input spec for singularity task
        a default value is used
    """
    filename = str(tmpdir.join("file_pydra.txt"))
    with open(filename, "w") as f:
        f.write("hello from pydra")

    cmd = "cat"
    image = "library://sylabsed/linux/alpine"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    default=filename,
                    metadata={"position": 1, "help_string": "input file"},
                ),
            )
        ],
        bases=(SingularitySpec,),
    )

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=cmd,
        input_spec=my_input_spec,
        strip=True,
        cache_dir=tmpdir,
    )

    res = singu()
    assert res.output.stdout == "hello from pydra"


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_inputspec_2(plugin, tmpdir):
    """ a customized input spec with two fields for singularity task """
    filename_1 = tmpdir.join("file_pydra.txt")
    with open(filename_1, "w") as f:
        f.write("hello from pydra\n")

    filename_2 = tmpdir.join("file_nice.txt")
    with open(filename_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    image = "library://sylabsed/linux/alpine"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file1",
                attr.ib(
                    type=File, metadata={"position": 1, "help_string": "input file 1"}
                ),
            ),
            (
                "file2",
                attr.ib(
                    type=File,
                    default=filename_2,
                    metadata={"position": 2, "help_string": "input file 2"},
                ),
            ),
        ],
        bases=(SingularitySpec,),
    )

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=cmd,
        file1=filename_1,
        input_spec=my_input_spec,
        strip=True,
        cache_dir=tmpdir,
    )

    res = singu()
    assert res.output.stdout == "hello from pydra\nhave a nice one"


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_inputspec_2a_except(plugin, tmpdir):
    """ a customized input spec with two fields
        first one uses a default, and second doesn't - raises a dataclass exception
    """
    filename_1 = tmpdir.join("file_pydra.txt")
    with open(filename_1, "w") as f:
        f.write("hello from pydra\n")
    filename_2 = tmpdir.join("file_nice.txt")
    with open(filename_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    image = "library://sylabsed/linux/alpine"

    # the field with default value can't be before value without default
    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file1",
                attr.ib(
                    type=File,
                    default=filename_1,
                    metadata={"position": 1, "help_string": "input file 1"},
                ),
            ),
            (
                "file2",
                attr.ib(
                    type=File, metadata={"position": 2, "help_string": "input file 2"}
                ),
            ),
        ],
        bases=(SingularitySpec,),
    )

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=cmd,
        file2=filename_2,
        input_spec=my_input_spec,
        strip=True,
        cache_dir=tmpdir,
    )
    res = singu()
    assert res.output.stdout == "hello from pydra\nhave a nice one"


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_inputspec_2a(plugin, tmpdir):
    """ a customized input spec with two fields
        first one uses a default value,
        this is fine even if the second field is not using any defaults
    """
    filename_1 = tmpdir.join("file_pydra.txt")
    with open(filename_1, "w") as f:
        f.write("hello from pydra\n")
    filename_2 = tmpdir.join("file_nice.txt")
    with open(filename_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    image = "library://sylabsed/linux/alpine"

    # if you want set default in the first field you can use default_value in metadata
    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file1",
                attr.ib(
                    type=File,
                    default=filename_1,
                    metadata={"position": 1, "help_string": "input file 1"},
                ),
            ),
            (
                "file2",
                attr.ib(
                    type=File, metadata={"position": 2, "help_string": "input file 2"}
                ),
            ),
        ],
        bases=(SingularitySpec,),
    )

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=cmd,
        file2=filename_2,
        input_spec=my_input_spec,
        strip=True,
        cache_dir=tmpdir,
    )

    res = singu()
    assert res.output.stdout == "hello from pydra\nhave a nice one"


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_cmd_inputspec_copyfile_1(plugin, tmpdir):
    """ shelltask changes a file in place,
        adding copyfile=True to the file-input from input_spec
        hardlink or copy in the output_dir should be created
    """
    file = tmpdir.join("file_pydra.txt")
    with open(file, "w") as f:
        f.write("hello from pydra\n")

    cmd = ["sed", "-is", "s/hello/hi/"]
    image = "library://sylabsed/linux/alpine"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "orig_file",
                attr.ib(
                    type=File,
                    metadata={
                        "position": 1,
                        "help_string": "orig file",
                        "mandatory": True,
                        "copyfile": True,
                    },
                ),
            ),
            (
                "out_file",
                attr.ib(
                    type=str,
                    metadata={
                        "output_file_template": "{orig_file}",
                        "help_string": "output file",
                    },
                ),
            ),
        ],
        bases=(SingularitySpec,),
    )

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=cmd,
        input_spec=my_input_spec,
        orig_file=str(file),
        cache_dir=tmpdir,
    )

    res = singu()
    assert res.output.stdout == ""
    assert res.output.out_file.exists()
    # the file is  copied, and than it is changed in place
    assert res.output.out_file.parent == singu.output_dir
    with open(res.output.out_file, "r") as f:
        assert "hi from pydra\n" == f.read()
    # the original file is unchanged
    with open(file, "r") as f:
        assert "hello from pydra\n" == f.read()


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_inputspec_state_1(plugin, tmpdir):
    """ a customised input spec for a singularity file with a splitter,
        splitter is on files
    """
    filename_1 = tmpdir.join("file_pydra.txt")
    with open(filename_1, "w") as f:
        f.write("hello from pydra\n")
    filename_2 = tmpdir.join("file_nice.txt")
    with open(filename_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    filename = [str(filename_1), str(filename_2)]
    image = "library://sylabsed/linux/alpine"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(SingularitySpec,),
    )

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=cmd,
        file=filename,
        input_spec=my_input_spec,
        strip=True,
        cache_dir=tmpdir,
    ).split("file")

    res = singu()
    assert res[0].output.stdout == "hello from pydra"
    assert res[1].output.stdout == "have a nice one"


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_inputspec_state_1b(plugin, tmpdir):
    """ a customised input spec for a singularity file with a splitter,
        files from the input spec have the same path in the local os and the container,
        so hash is calculated and the test works fine
    """
    file_1 = tmpdir.join("file_pydra.txt")
    file_2 = tmpdir.join("file_nice.txt")
    with open(file_1, "w") as f:
        f.write("hello from pydra")
    with open(file_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    filename = [str(file_1), str(file_2)]
    image = "library://sylabsed/linux/alpine"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(SingularitySpec,),
    )

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=cmd,
        file=filename,
        input_spec=my_input_spec,
        strip=True,
        cache_dir=tmpdir,
    ).split("file")

    res = singu()
    assert res[0].output.stdout == "hello from pydra"
    assert res[1].output.stdout == "have a nice one"


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_wf_inputspec_1(plugin, tmpdir):
    """ a customized input spec for workflow with singularity tasks """
    filename = tmpdir.join("file_pydra.txt")
    with open(filename, "w") as f:
        f.write("hello from pydra")

    cmd = "cat"
    image = "library://sylabsed/linux/alpine"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(SingularitySpec,),
    )

    wf = Workflow(name="wf", input_spec=["cmd", "file"], cache_dir=tmpdir)
    wf.inputs.cmd = cmd
    wf.inputs.file = filename

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=wf.lzin.cmd,
        file=wf.lzin.file,
        input_spec=my_input_spec,
        strip=True,
    )
    wf.add(singu)

    wf.set_output([("out", wf.singu.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == "hello from pydra"


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_wf_state_inputspec_1(plugin, tmpdir):
    """ a customized input spec for workflow with singularity tasks that has a state"""
    file_1 = tmpdir.join("file_pydra.txt")
    file_2 = tmpdir.join("file_nice.txt")
    with open(file_1, "w") as f:
        f.write("hello from pydra")
    with open(file_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    filename = [str(file_1), str(file_2)]
    image = "library://sylabsed/linux/alpine"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(SingularitySpec,),
    )

    wf = Workflow(name="wf", input_spec=["cmd", "file"], cache_dir=tmpdir)
    wf.inputs.cmd = cmd
    wf.inputs.file = filename

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=wf.lzin.cmd,
        file=wf.lzin.file,
        input_spec=my_input_spec,
        strip=True,
    )
    wf.add(singu)
    wf.split("file")

    wf.set_output([("out", wf.singu.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res[0].output.out == "hello from pydra"
    assert res[1].output.out == "have a nice one"


@need_singularity
@pytest.mark.parametrize("plugin", Plugins)
def test_singularity_wf_ndst_inputspec_1(plugin, tmpdir):
    """ a customized input spec for workflow with singularity tasks with states"""
    file_1 = tmpdir.join("file_pydra.txt")
    file_2 = tmpdir.join("file_nice.txt")
    with open(file_1, "w") as f:
        f.write("hello from pydra")
    with open(file_2, "w") as f:
        f.write("have a nice one")

    cmd = "cat"
    filename = [str(file_1), str(file_2)]
    image = "library://sylabsed/linux/alpine"

    my_input_spec = SpecInfo(
        name="Input",
        fields=[
            (
                "file",
                attr.ib(
                    type=File,
                    metadata={
                        "mandatory": True,
                        "position": 1,
                        "help_string": "input file",
                    },
                ),
            )
        ],
        bases=(SingularitySpec,),
    )

    wf = Workflow(name="wf", input_spec=["cmd", "file"], cache_dir=tmpdir)
    wf.inputs.cmd = cmd
    wf.inputs.file = filename

    singu = SingularityTask(
        name="singu",
        image=image,
        executable=wf.lzin.cmd,
        file=wf.lzin.file,
        input_spec=my_input_spec,
        strip=True,
    ).split("file")
    wf.add(singu)

    wf.set_output([("out", wf.singu.lzout.stdout)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.out == ["hello from pydra", "have a nice one"]



# Tasks for testing
import time
import typing as tp
from pathlib import Path

from ..core import Workflow
from ... import mark
from ..specs import File


@mark.task
def fun_addtwo(a):
    import time

    time.sleep(1)
    if a == 3:
        time.sleep(2)
    return a + 2


@mark.task
def fun_addvar(a, b):
    return a + b


@mark.task
def fun_addvar_none(a, b):
    if b is None:
        return a
    else:
        return a + b


@mark.task
def fun_addvar_default(a, b=1):
    return a + b


@mark.task
def fun_addvar3(a, b, c):
    return a + b + c


@mark.task
def fun_addvar4(a, b, c, d):
    return a + b + c + d


@mark.task
def moment(lst, n):
    return sum([i ** n for i in lst]) / len(lst)


@mark.task
def fun_div(a, b):
    return a / b


@mark.task
def multiply(x, y):
    return x * y


@mark.task
def add2(x):
    if x == 1 or x == 12:
        time.sleep(1)
    return x + 2


@mark.task
@mark.annotate({"return": {"out_add": float, "out_sub": float}})
def add2_sub2_res(res):
    """function that takes entire output as an input"""
    if isinstance(res, list):
        return [r["out"] + 2 for r in res], [r["out"] - 2 for r in res]
    return res["out"] + 2, res["out"] - 2


@mark.task
def power(a, b):
    return a ** b


@mark.task
def identity(x):
    return x


@mark.task
def add2_wait(x):
    time.sleep(3)
    return x + 2


@mark.task
def list_output(x):
    return [x, 2 * x, 3 * x]


@mark.task
def fun_dict(d):
    kv_list = [f"{k}:{v}" for (k, v) in d.items()]
    return "_".join(kv_list)


@mark.task
def fun_write_file(filename: tp.Union[str, File, Path], text="hello"):
    with open(filename, "w") as f:
        f.write(text)
    return Path(filename).absolute()


@mark.task
def fun_write_file_list(filename_list: tp.List[tp.Union[str, File, Path]], text="hi"):
    for ii, filename in enumerate(filename_list):
        with open(filename, "w") as f:
            f.write(f"from file {ii}: {text}")
    filename_list = [Path(filename).absolute() for filename in filename_list]
    return filename_list


@mark.task
def fun_write_file_list2dict(
    filename_list: tp.List[tp.Union[str, File, Path]], text="hi"
):
    filename_dict = {}
    for ii, filename in enumerate(filename_list):
        with open(filename, "w") as f:
            f.write(f"from file {ii}: {text}")
        filename_dict[f"file_{ii}"] = Path(filename).absolute()
    # adding an additional field with int
    filename_dict["random_int"] = 20
    return filename_dict


@mark.task
def fun_file(filename: File):
    with open(filename) as f:
        txt = f.read()
    return txt


@mark.task
def fun_file_list(filename_list: File):
    txt_list = []
    for filename in filename_list:
        with open(filename) as f:
            txt_list.append(f.read())
    return " ".join(txt_list)


def gen_basic_wf(name="basic-wf"):
    """
    Generates `Workflow` of two tasks

    Task Input
    ----------
    x : int (5)

    Task Output
    -----------
    out : int (9)
    """
    wf = Workflow(name=name, input_spec=["x"])
    wf.inputs.x = 5
    wf.add(fun_addtwo(name="task1", a=wf.lzin.x, b=0))
    wf.add(fun_addvar(name="task2", a=wf.task1.lzout.out, b=2))
    wf.set_output([("out", wf.task2.lzout.out)])
    return wf



"""Module to keep track of provenance information."""
import os
from pathlib import Path
import json
import attr
from ..utils.messenger import send_message, make_message, gen_uuid, now, AuditFlag
from .helpers import ensure_list, gather_runtime_info


class Audit:
    """Handle provenance tracking and resource utilization."""

    def __init__(self, audit_flags, messengers, messenger_args, develop=None):
        """
        Initialize the auditing functionality.

        Parameters
        ----------
        audit_flags : :class:`AuditFlag`
            Base configuration of auditing.
        messengers :
            TODO
        messenger_args :
            TODO
        develop :
            TODO

        """
        self.audit_flags = audit_flags
        self.messengers = ensure_list(messengers)
        self.messenger_args = messenger_args
        self.develop = develop

    def start_audit(self, odir):
        """
        Start recording provenance.

        Monitored information is not sent until directory is created,
        in case message directory is inside task output directory.

        Parameters
        ----------
        odir : :obj:`os.pathlike`
            Message output directory.

        """
        self.odir = odir
        if self.audit_check(AuditFlag.PROV):
            self.aid = "uid:{}".format(gen_uuid())
            start_message = {"@id": self.aid, "@type": "task", "startedAtTime": now()}
        os.chdir(self.odir)
        if self.audit_check(AuditFlag.PROV):
            self.audit_message(start_message, AuditFlag.PROV)
        if self.audit_check(AuditFlag.RESOURCE):
            from ..utils.profiler import ResourceMonitor

            self.resource_monitor = ResourceMonitor(os.getpid(), logdir=self.odir)

    def monitor(self):
        """Start resource monitoring."""
        if self.audit_check(AuditFlag.RESOURCE):
            self.resource_monitor.start()
            if self.audit_check(AuditFlag.PROV):
                self.mid = "uid:{}".format(gen_uuid())
                self.audit_message(
                    {
                        "@id": self.mid,
                        "@type": "monitor",
                        "startedAtTime": now(),
                        "wasStartedBy": self.aid,
                    },
                    AuditFlag.PROV,
                )

    def finalize_audit(self, result):
        """End auditing."""
        if self.audit_check(AuditFlag.RESOURCE):
            self.resource_monitor.stop()
            result.runtime = gather_runtime_info(self.resource_monitor.fname)
            if self.audit_check(AuditFlag.PROV):
                self.audit_message(
                    {"@id": self.mid, "endedAtTime": now(), "wasEndedBy": self.aid},
                    AuditFlag.PROV,
                )
                # audit resources/runtime information
                self.eid = "uid:{}".format(gen_uuid())
                entity = attr.asdict(result.runtime)
                entity.update(
                    **{
                        "@id": self.eid,
                        "@type": "runtime",
                        "prov:wasGeneratedBy": self.aid,
                    }
                )
                self.audit_message(entity, AuditFlag.PROV)
                self.audit_message(
                    {
                        "@type": "prov:Generation",
                        "entity_generated": self.eid,
                        "hadActivity": self.mid,
                    },
                    AuditFlag.PROV,
                )
            self.resource_monitor = None
        if self.audit_check(AuditFlag.PROV):
            # audit outputs
            self.audit_message(
                {"@id": self.aid, "endedAtTime": now(), "errored": result.errored},
                AuditFlag.PROV,
            )

    def audit_message(self, message, flags=None):
        """
        Send auditing message.

        Parameters
        ----------
        message :
            TODO
        flags :
            TODO

        """
        if self.develop:
            with open(
                Path(os.path.dirname(__file__)) / ".." / "schema/context.jsonld", "rt"
            ) as fp:
                context = json.load(fp)
        else:
            context = {
                "@context": "https://raw.githubusercontent.com/nipype/pydra"
                "/master/pydra/schema/context.jsonld"
            }
        if self.audit_flags & flags:
            if self.messenger_args:
                send_message(
                    make_message(message, context=context),
                    messengers=self.messengers,
                    **self.messenger_args,
                )
            else:
                send_message(
                    make_message(message, context=context), messengers=self.messengers
                )

    def audit_check(self, flag):
        """
        Determine whether auditing is enabled for a particular flag.

        Parameters
        ----------
        flag :
            TODO

        """
        return self.audit_flags & flag



"""Messaging of states."""
import abc
import datetime as dt
import enum
from pathlib import Path
import os


def gen_uuid():
    """Generate a unique identifier."""
    import uuid

    return uuid.uuid4().hex


def now():
    """Get a formatted timestamp."""
    return dt.datetime.utcnow().isoformat(timespec="microseconds")


class AuditFlag(enum.Flag):
    """Auditing flags."""

    NONE = 0
    """Do not track provenance or monitor resources."""
    PROV = enum.auto()  # 0x01
    """Track provenance only."""
    RESOURCE = enum.auto()  # 0x02
    """Monitor resource utilization only."""
    ALL = PROV | RESOURCE
    """Track provenance and resource utilization."""


class RuntimeHooks(enum.IntEnum):
    """Allowed points to hook into the process."""

    # TODO: Convert all messenger actions to hooks at the following stages
    task_run_entry = enum.auto()
    task_run_exit = enum.auto()
    resource_monitor_pre_start = enum.auto()
    resource_monitor_post_stop = enum.auto()
    task_execute_pre_entry = enum.auto()
    task_execute_post_exit = enum.auto()


class Messenger:
    """Base messenger class."""

    @abc.abstractmethod
    def send(self, message, **kwargs):
        """Send a message."""


class PrintMessenger(Messenger):
    """A messenger that redirects to standard output."""

    def send(self, message, **kwargs):
        """
        Send to standard output.

        Parameters
        ----------
        message : :obj:`dict`
            The message to be printed.

        """
        import json

        mid = gen_uuid()
        print(
            "id: {0}\n{1}".format(
                mid, json.dumps(message, ensure_ascii=False, indent=2, sort_keys=False)
            )
        )


class FileMessenger(Messenger):
    """A messenger that redirects to a file."""

    def send(self, message, append=True, **kwargs):
        """
        Append message to file.

        Parameters
        ----------
        message : :obj:`dict`
            The message to be printed.
        append : :obj:`bool`
            Do not truncate file when opening (i.e. append to it).

        """
        import json

        mid = gen_uuid()
        if kwargs.get("message_dir"):
            message_dir = Path(kwargs["message_dir"])
        else:
            message_dir = Path(os.getcwd()) / "messages"
        message_dir.mkdir(parents=True, exist_ok=True)
        with (message_dir / ("%s.jsonld" % mid)).open("wa"[append]) as fp:
            json.dump(message, fp, ensure_ascii=False, indent=2, sort_keys=False)
        return mid


class RemoteRESTMessenger(Messenger):
    """A messenger that redirects to remote REST endpoint."""

    def send(self, message, **kwargs):
        """
        Append message to file.

        Parameters
        ----------
        message : :obj:`dict`
            The message to be printed.

        """
        import requests

        r = requests.post(
            kwargs["post_url"],
            json=message,
            auth=kwargs["auth"]()
            if getattr(kwargs["auth"], "__call__", None)
            else kwargs["auth"],
        )
        return r.status_code


def send_message(message, messengers=None, **kwargs):
    """Send NIDM messages for logging provenance and auditing."""
    for messenger in messengers:
        messenger.send(message, **kwargs)


def make_message(obj, context=None):
    """Build a message."""
    if context is None:
        context = {
            "@context": "https://raw.githubusercontent.com/"
            "nipype/pydra/enh/task/pydra/schema/context.jsonld"
        }
    message = context.copy()
    message.update(**obj)
    return message


def collect_messages(collected_path, message_path, ld_op="compact"):
    """Gather messages."""
    import pyld as pld
    import json
    from glob import glob

    fl = glob(str(message_path / "*.jsonld"))
    data = []
    for f in fl:
        with open(f, "rt") as fp:
            data.append(json.load(fp))
    if data:
        records = getattr(pld.jsonld, ld_op)(
            pld.jsonld.from_rdf(pld.jsonld.to_rdf(data, {})), data[0]
        )
        records["@id"] = "uid:{}".format(gen_uuid())
        with open(collected_path / "messages.jsonld", "wt") as fp:
            json.dump(records, fp, ensure_ascii=False, indent=2, sort_keys=False)
